# -*- coding: utf-8 -*-
import json
import os
import shutil
import tempfile

import tensorflow as tf
from tensorflow.keras.layers import Embedding

from sknlp.vocab import Vocab
from sknlp.utils import make_tarball


class Token2vec:

    def __init__(self, vocab, embed_size, segmenter='list',
                 name='token2vec', **kwargs):
        """
        基础符号->向量模块.

        Parameters:
        ----------
        vocab: sknlp.vocab.Vocab. 符号表.
        embed_size: int > 0. 向量维度.
        name: str. 模块名.
        embeddings_initializer: Initializer for the `embeddings` matrix.
        embeddings_regularizer: Regularizer function applied to
        the `embeddings` matrix.
        embeddings_constraint: Constraint function applied to
        the `embeddings` matrix.
        input_length: Length of input sequences, when it is constant.
        This argument is required if you are going to connect
        `Flatten` then `Dense` layers upstream
        (without it, the shape of the dense outputs cannot be computed).

        Inputs
        ----------
        2D tensor with shape: `(batch_size, input_length)`.

        Outputs
        ----------
        3D tensor with shape: `(batch_size, input_length, embed_size)`.
        """
        self._vocab = vocab
        self._embedding = Embedding(len(vocab), embed_size, mask_zero=True,
                                    name='embeddings', **kwargs)
        self._model = tf.keras.Sequential(self._embedding, name=name)
        self._embed_size = embed_size
        self._segmenter = segmenter
        self._kwargs = kwargs

    def __call__(self, inputs):
        return self._embedding(inputs)

    def compute_mask(self, inputs):
        return self._embedding.compute_mask(inputs)

    def save(self, filename):
        with tempfile.TemporaryDirectory() as temp_dir:
            self._model.save_weights(os.path.join(temp_dir, 'weights.h5'))
            with open(os.path.join(temp_dir, 'vocab.json'), 'w') as f:
                f.write(self._vocab.to_json())
            with open(os.path.join(temp_dir, 'meta.json'), 'w') as f:
                f.write(json.dumps(self.get_config()))

            if not filename.endswith('.tar'):
                filename = '.'.join([filename, 'tar'])
            make_tarball(filename, temp_dir)

    @classmethod
    def load(cls, filename):
        with tempfile.TemporaryDirectory() as temp_dir:
            shutil.unpack_archive(filename, temp_dir, format='tar')
            with open(os.path.join(temp_dir, 'vocab.json')) as f:
                vocab = Vocab.from_json(f.read())
            with open(os.path.join(temp_dir, 'meta.json')) as f:
                meta = json.loads(f.read())

            module = cls(vocab, **meta)
            module._model.load_weights(os.path.join(temp_dir, 'weights.h5'))
            return module

    def freeze(self):
        for layer in self._model.layers:
            layer.trainable = False

    def get_config(self):
        return {'embed_size': self._embed_size,
                'segmenter': self._segmenter,
                **self._kwargs}

    @property
    def vocab(self):
        return self._vocab

    @property
    def segmenter(self):
        return self._segmenter

    @property
    def embed_size(self):
        return self._embed_size

    @property
    def input(self):
        return self._model.input

    @property
    def output(self):
        return self._model.output

    @property
    def input_shape(self):
        return self._model.input.shape

    @property
    def output_shape(self):
        return self._model.output.shape

    @property
    def weights(self):
        return self._model.get_weights()

    def set_weights(self, weights):
        return self._model.set_weights(weights)



from typing import Sequence, Union, Optional, Dict, Any

import json
import os
import itertools
from collections import Counter
import logging

import tensorflow as tf
import pandas as pd

from sknlp.data import NLPDataset
from sknlp.data.text_segmenter import get_segmenter
from sknlp.vocab import Vocab
from .embedding import Token2vec

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


class BaseNLPModel:

    def __init__(
        self,
        segmenter="jieba",
        embed_size=100,
        max_length=None,
        vocab=None,
        token2vec=None
    ):
        self._max_length = max_length
        self._token2vec = token2vec
        if token2vec is not None:
            self._segmenter = token2vec.segmenter
            self._vocab = token2vec.vocab
            self._embed_size = token2vec.embed_size
        else:
            self._vocab = vocab
            self._segmenter = segmenter
            self._embed_size = embed_size
        self._built = False

    @staticmethod
    def build_vocab(texts, segment_func, min_frequency=5):
        counter = Counter(
            itertools.chain.from_iterable(segment_func(text) for text in texts)
        )
        return Vocab(counter, min_frequency=min_frequency)

    def build_token2vec(self, vocab):
        self._vocab = vocab
        self._token2vec = Token2vec(vocab, self._embed_size, self._segmenter)

    def build_encode_layer(self, inputs):
        raise NotImplementedError()

    def build_output_layer(self, inputs):
        raise NotImplementedError()

    def build(self):
        if self._token2vec is None:
            self.build_token2vec(self._vocab)
        inputs = self._token2vec.output
        self._model = tf.keras.Model(
            inputs=self._token2vec.input,
            outputs=self.build_output_layer(self.build_encode_layer(inputs))
        )
        self._built = True

    def get_loss(self):
        raise NotImplementedError()

    def get_metrics(self):
        raise NotImplementedError()

    def get_callbacks(self):
        raise NotImplementedError()

    def get_custom_objects(self):
        return {}

    def save_vocab(self, directory, filename="vocab.json"):
        with open(os.path.join(directory, filename), "w") as f:
            f.write(self._vocab.to_json())

    def save_config(self, directory, filename="meta.json"):
        with open(os.path.join(directory, filename), "w") as f:
            f.write(json.dumps(self.get_config()))

    def save(self, directory):
        self._model.save(os.path.join(directory, "model"), save_format="h5")
        self.save_vocab(directory)
        self.save_config(directory)

    @classmethod
    def load(cls, directory):
        with open(os.path.join(directory, "vocab.json")) as f:
            vocab = Vocab.from_json(f.read())
        with open(os.path.join(directory, "meta.json")) as f:
            meta = json.loads(f.read())
        meta.pop("algorithm", None)
        meta.pop("task", None)
        module = cls(vocab=vocab, **meta)
        module._model = tf.keras.models.load_model(
            os.path.join(directory, "model"),
            custom_objects=module.get_custom_objects()
        )
        module._built = True
        return module

    def export(self, directory, name, version="0"):
        d = os.path.join(directory, name, version)
        self._model.save(d, save_format="tf")
        self.save_vocab(d)
        self.save_config(d)

    def get_config(self):
        return {
            "segmenter": self._segmenter,
            "embed_size": self._embed_size,
            "max_length": self._max_length
        }


class SupervisedNLPModel(BaseNLPModel):

    def __init__(
        self,
        classes: Sequence[str],
        segmenter: str = "jieba",
        embed_size: int = 100,
        max_length: Optional[int] = 100,
        vocab: Optional[Vocab] = None,
        token2vec: Optional[Token2vec] = None,
        task: Optional[str] = None,
        algorithm: Optional[str] = None,
        **kwargs
    ):
        super().__init__(segmenter=segmenter,
                         embed_size=embed_size,
                         max_length=max_length,
                         vocab=vocab,
                         token2vec=token2vec,
                         **kwargs)
        self._task = task
        self._algorithm = algorithm
        self._class2idx = dict(zip(list(classes), range(len(classes))))
        self._idx2class = dict(zip(range(len(classes)), list(classes)))
        self._num_classes = len(classes)

    @classmethod
    def create_dataset_from_df(
        self,
        df: pd.DataFrame,
        vocab: Vocab,
        labels: Sequence[str]
    ) -> NLPDataset:
        raise NotImplementedError()

    def prepare_tf_dataset(
        self,
        X: Sequence[str],
        y: Sequence[str],
        dataset: NLPDataset,
        batch_size: int,
        shuffle: bool = True
    ) -> tf.data.Dataset:
        assert not ((X is None or y is None) and dataset is None)

        if self._vocab is None:
            cut = get_segmenter(self._segmenter)
            self._vocab = self.build_vocab(X, cut)
        if dataset is not None:
            return dataset.batchify(batch_size, shuffle=shuffle)
        if isinstance(y[0], (list, tuple)):
            y = ["|".join(map(str, y_i)) for y_i in y]
        df = pd.DataFrame(zip(X, y), columns=["text", "label"])
        dataset = self.create_dataset_from_df(df, self._vocab, self._class2idx.keys())
        return dataset.batchify(
            batch_size, shuffle=shuffle, shuffle_buffer_size=df.shape[0]
        )

    def fit(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: NLPDataset = None,
        valid_X: Sequence[str] = None,
        valid_y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        valid_dataset: NLPDataset = None,
        batch_size: int = 128,
        n_epochs: int = 30,
        optimizer: str = "adam",
        lr: float = 1e-3,
        lr_update_factor: float = 0.5,
        lr_update_epochs: int = 10,
        clip: float = 5.0,
        checkpoint: Optional[str] = None,
        save_frequency: int = 1,
        log_file: Optional[str] = None,
        verbose: int = 2
    ) -> None:
        assert not (self._token2vec is None
                    and self._vocab is None
                    and X is None), \
            "When token2vec and vocab are both not given, X must be provided"

        self.train_dataset = self.prepare_tf_dataset(X, y, dataset, batch_size)
        if ((valid_X is None or valid_y is None) and valid_dataset is None):
            self.valid_dataset = None
        else:
            self.valid_dataset = self.prepare_tf_dataset(
                valid_X, valid_y, valid_dataset, batch_size, shuffle=False
            )
        if not self._built:
            self.build()
        optimizer = tf.keras.optimizers.Adam(learning_rate=lr, clipnorm=clip)
        if optimizer != "adam":
            pass
        loss = self.get_loss()
        metrics = self.get_metrics()
        callbacks = self.get_callbacks()
        self._model.compile(optimizer=optimizer, loss=loss, metrics=metrics)

        def scheduler(epoch):
            new_lr = lr * lr_update_factor ** (epoch // lr_update_epochs)
            if epoch != 0 and epoch % lr_update_epochs == 0:
                logger.info("change learning rate to %.4g" % new_lr)
            return new_lr

        lr_decay = tf.keras.callbacks.LearningRateScheduler(scheduler)
        callbacks = [*callbacks, lr_decay]
        if log_file is not None:
            callbacks.append(tf.keras.callbacks.CSVLogger(log_file))
        self._model.fit(
            self.train_dataset,
            epochs=n_epochs,
            validation_data=self.valid_dataset,
            callbacks=callbacks,
            verbose=verbose
        )

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: NLPDataset = None,
        batch_size: int = 128
    ) -> tf.Tensor:
        assert self._built
        y = None if X is None else ["O" for _ in range(len(X))]
        dataset = self.prepare_tf_dataset(X, y, dataset, batch_size, shuffle=False)

        def _f(x, y):
            return x

        return self._model.predict(dataset.map(_f))

    def score(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: NLPDataset = None,
        thresholds: Union[float, Dict[str, float]] = 0.5,
        batch_size: int = 128
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "classes": list(self._class2idx.keys()),
            "task": self._task,
            "algorithm": self._algorithm
        }



import json
from collections import defaultdict, Counter


class Vocab:

    def __init__(self, counter=None, min_frequency=1,
                 unk_token='<unk>', padding_token='<pad>',
                 bos_token='<bos>', eos_token='<eos>'):
        self._min_frequency = min_frequency
        self._unk_token = unk_token
        self._padding_token = padding_token
        self._bos_token = bos_token
        self._eos_token = eos_token

        self._token2idx = defaultdict(lambda: 1)
        self._token2idx[padding_token] = 0
        self._token2idx[unk_token] = 1
        self._token2idx[bos_token] = 2
        self._token2idx[eos_token] = 3
        reversed_tokens = {unk_token, padding_token, bos_token, eos_token}

        self._token_frequency = defaultdict(lambda: 0)
        if counter is not None:
            for _, (token, frequency) in enumerate(counter.most_common()):
                if token not in reversed_tokens and frequency >= min_frequency:
                    self._token2idx[token] = len(self._token2idx)
                    self._token_frequency[token] = frequency
        self._idx2token = dict(zip(self._token2idx.values(),
                                   self._token2idx.keys()))

    def idx2token(self, indices):
        """
        Lookup tokens by indices.

        raise KeyError if any indices are greater than or equal to
        size of vacab.

        Parameters
        ----------
        indices: int or list of int

        Returns
        ----------
        str or list of str
        """
        if not isinstance(indices, (list, tuple)):
            if indices not in self._idx2token:
                raise(KeyError('index %d is out of vacab' % indices))
            return self._idx2token[indices]
        else:
            res = []
            for idx in indices:
                if idx not in self._idx2token:
                    raise(KeyError('index %d is out of vacab' % idx))
                res.append(self.idx2token(idx))
            return res

    def token2idx(self, tokens):
        if not isinstance(tokens, (list, tuple)):
            return self._token2idx[tokens]
        else:
            return [self.token2idx(t) for t in tokens]

    def to_json(self):
        return json.dumps(
            {
                'tokens': self._token_frequency,
                'token2idx': self._token2idx,
                'unk': self._unk_token,
                'pad': self._padding_token,
                'bos': self._bos_token,
                'eos': self._eos_token
            },
            ensure_ascii=False
        )

    @classmethod
    def from_json(cls, json_str):
        vocab_dict = json.loads(json_str)
        return cls(Counter(vocab_dict['tokens']),
                   min_frequency=1,
                   unk_token=vocab_dict['unk'],
                   padding_token=vocab_dict['pad'],
                   bos_token=vocab_dict['bos'],
                   eos_token=vocab_dict['eos'])

    def __getitem__(self, tokens):
        return self.token2idx(tokens)

    def __contains__(self, token):
        return token in self._token2idx

    def __len__(self):
        return len(self._token2idx)

    def __repr__(self):
        return 'Vocab(size=%d, unk="%s")' % (len(self), self._unk_token)

    @property
    def pad(self):
        return self._padding_token

    @property
    def unk(self):
        return self._unk_token

    @property
    def bos(self):
        return self._bos_token

    @property
    def eos(self):
        return self._eos_token
